@@ -1,5 +1,6 @@
 import os
-from typing import List
+import re
+from typing import Set
 from shutil import rmtree, move
 import sqlite3
 import json
@@ -8,9 +9,10 @@ from pathlib import Path
 import fire  # type: ignore
 from tqdm import tqdm  # type: ignore
 import rpyc  # type: ignore
+from pywebcopy import WebPage, config  # type: ignore
 from .persistor import Persistor
 from .graph import graph
-from .utils import remove_symlink
+from .utils import remove_dir, remove_symlink, download_url
 
 
 class HyperTag:
@@ -50,7 +52,7 @@ class HyperTag:
         """ Vectorize image files (needed for semantic search) """
         from .vectorizer import CLIPVectorizer, get_image_files
 
-        file_paths = self.db.get_unindexed_file_paths()
+        file_paths = self.db.get_unvectorized_file_paths()
         compatible_files = get_image_files(file_paths, verbose=True)
         print("Vectorizing", len(compatible_files), "images...")
         remote = True
@@ -74,6 +76,12 @@ class HyperTag:
 
             self.db.add_file_embedding_vector(file_path, json.dumps(img_vector))
             self.db.conn.commit()
+        print("Updating index...")
+        if remote:
+            rpc.root.update_image_index()
+        else:
+            corpus_vectors, corpus_paths = img_vectorizer.get_image_corpus()
+            img_vectorizer.update_index(len(corpus_vectors))
 
     def index_texts(self, rebuild=False, cache=False, cores: int = 0):
         """ Vectorize text files (needed for semantic search) """
@@ -95,9 +103,9 @@ class HyperTag:
             print("Caching cleaned texts (database will grow big)")
         if rebuild:
             print("Rebuilding index")
-            file_paths = self.db.get_indexed_file_paths()
+            file_paths = self.db.get_vectorized_file_paths()
         else:
-            file_paths = self.db.get_unindexed_file_paths()
+            file_paths = self.db.get_unvectorized_file_paths()
         i = 0
         compatible_files = get_text_documents(file_paths, verbose=True)
         min_words = 5
@@ -143,6 +151,12 @@ class HyperTag:
                 self.db.conn.commit()
                 print("Failed to parse file - skipping:", file_path)
         print(f"Vectorized {str(i)} file/s successfully")
+        print("Updating index...")
+        if remote:
+            rpc.root.update_text_index()
+        else:
+            corpus_vectors, corpus_paths = vectorizer.get_text_corpus()
+            vectorizer.update_index(len(corpus_vectors))
 
     def search(self, *text_queries: str, path=False, top_k=10, score=False, _return=False):
         """ Execute a semantic search that returns best matching text documents """
@@ -162,9 +176,9 @@ class HyperTag:
         if _return:
             return results
 
-    def add_auto_import_dir(self, path: str):
+    def add_auto_import_dir(self, path: str, index_images=False, index_texts=False):
         """ Add path for auto import directory (watched by daemon) """
-        self.db.add_auto_import_directory(path)
+        self.db.add_auto_import_directory(path, index_images, index_texts)
 
     def set_hypertagfs_dir(self, path: str):
         """ Set path for HyperTagFS directory """
@@ -182,13 +196,21 @@ class HyperTag:
             os.makedirs(root_path / "Search Texts", exist_ok=True)
             os.makedirs(root_path / "Search Images", exist_ok=True)
             tag_ids_names = self.db.get_root_tag_ids_names()
+            parent_file_paths_names = None
         else:
             tag_ids_names = self.db.get_tag_id_children_ids_names(parent_tag_id)
+            parent_file_paths_names = set(
+                self.db.get_file_paths_names_by_tag_id_shallow(parent_tag_id)
+            )
 
         leaf_tag_ids = {tag_id[0] for tag_id in self.db.get_leaf_tag_ids()}
-
+        dupes = dict()
         for tag_id, name in tag_ids_names:
-            file_paths_names = self.db.get_file_paths_names_by_tag_id(tag_id)
+            child_file_paths_names = set(self.db.get_file_paths_names_by_tag_id_shallow(tag_id))
+            if parent_file_paths_names is None:
+                file_paths_names = child_file_paths_names
+            else:  # Intersect parent files with child
+                file_paths_names = parent_file_paths_names.intersection(child_file_paths_names)
             if len(file_paths_names) > 0:
                 underscore_root_tag_path = root_path / ("_" + name)
                 root_tag_path = root_path / name
@@ -201,13 +223,25 @@ class HyperTag:
                     os.makedirs(symlink_path, exist_ok=True)
                 for file_path, file_name in file_paths_names:
                     try:
-                        os.symlink(Path(file_path), symlink_path / file_name)
+                        filepath = Path(file_path)
+                        current_symlink_path = symlink_path / file_name
+                        if current_symlink_path.exists() and current_symlink_path.is_symlink():
+                            existing_target = current_symlink_path.resolve()  # Get symlink target
+                            if str(existing_target) != str(filepath):  # Duplicate?
+                                dupe_i = dupes.get(current_symlink_path)
+                                if dupe_i is None:
+                                    dupes[current_symlink_path] = 1
+                                    dupe_i = 2
+                                dupes[current_symlink_path] += 1
+                                current_symlink_path = symlink_path / (f"{dupe_i}-" + file_name)
+
+                        os.symlink(filepath, current_symlink_path)
                     except FileExistsError:
                         pass
                 self.mount(root_tag_path, tag_id)
 
     def auto_add_tags_from_path(
-        self, file_path: str, import_path_dirs: List[str], verbose=False, keep_all=False
+        self, file_path: Path, import_path_dirs: Set[str], verbose=False, keep_all=False
     ):
         file_path_tags = [p for p in str(file_path).split("/") if p not in import_path_dirs]
         if not keep_all:
@@ -215,7 +249,7 @@ class HyperTag:
         if verbose:
             print("Inferred tags:", file_path_tags)
         self.tag(
-            file_path.name,
+            file_path,
             "with",
             *file_path_tags,
             remount=False,
@@ -259,14 +293,47 @@ class HyperTag:
             remove_symlink(self.root_dir, file_name)
         self.mount(self.root_dir)
 
-    def add(self, *file_paths):
-        """ Add file/s """
+    def scrape(self, url, folder, timeout=1):
+        config.setup_config(url, folder)
+        wp = WebPage()
+        wp.get(url)
+        # start the saving process
+        wp.save_complete()
+        # join the sub threads
+        for t in wp._threads:
+            if t.is_alive():
+                t.join(timeout)
+        # location of the html file written
+        return wp.file_path
+
+    def add_url(self, url):
+        webpages_regex = r"\S+.html|\S+.htm|\S+.php|\S+\/[^.]+"
+        matches = re.findall(webpages_regex, url)
+        if len(matches[0]) == len(url):
+            web_pages_path = self.db.db_path / "web_pages"
+            os.makedirs(web_pages_path, exist_ok=True)
+            index_path = self.scrape(url, str(web_pages_path))
+            return index_path
+        else:
+            downloads_path = self.db.db_path / "downloads"
+            os.makedirs(downloads_path, exist_ok=True)
+            file_name = url.split("/")[-1]
+            file_path = downloads_path / file_name
+            download_url(url, file_path)
+            return file_path
+
+    def add(self, *paths):
+        """ Add file/s or URL/s"""
         added = []
-        for file_path in tqdm(file_paths):
+        for path in tqdm(paths):
             try:
-                if Path(file_path).is_file():
+                if str(path).startswith("http"):
+                    file_path = self.add_url(path)
                     self.db.add_file(os.path.abspath(file_path))
-                    added.append(file_path)
+                    added.append(path)
+                elif Path(path).is_file():
+                    self.db.add_file(os.path.abspath(path))
+                    added.append(path)
             except sqlite3.IntegrityError:
                 pass
         self.db.conn.commit()
@@ -286,7 +353,7 @@ class HyperTag:
         if mode == "files":
             names = self.db.get_files(path)
         elif mode == "index":
-            names = self.db.get_indexed_file_paths(path)
+            names = self.db.get_vectorized_file_paths(path)
         elif mode == "tags":
             names = self.db.get_tags()
         for name in names:
@@ -301,11 +368,23 @@ class HyperTag:
         """
         # TODO: Parse AST to support queries with brackets
         operands = {"and", "or", "minus"}
-        results = set(self.db.get_files_by_tag(query[0], path, fuzzy, verbose))
+        tag_val = query[0].split("=")
+        if len(tag_val) == 1:
+            tag_query, value = (query[0], None)
+        else:
+            tag_query, value = (tag_val[0], tag_val[-1])
+            value = value.replace("*", "%")
+        results = set(self.db.get_files_by_tag(tag_query, path, fuzzy, value, verbose))
         current_operand = None
         for query_symbol in query[1:]:
             if query_symbol not in operands:
-                file_names = set(self.db.get_files_by_tag(query_symbol, path, fuzzy, verbose))
+                tag_val = query_symbol.split("=")
+                if len(tag_val) == 1:
+                    tag_query, value = (query_symbol, None)
+                else:
+                    tag_query, value = (tag_val[0], tag_val[-1])
+                    value = value.replace("*", "%")
+                file_names = set(self.db.get_files_by_tag(tag_query, path, fuzzy, value, verbose))
                 if current_operand == "or":
                     results = results.union(file_names)
                 elif current_operand == "minus":
@@ -320,25 +399,31 @@ class HyperTag:
     def tag(self, *args, remount=True, add=True, commit=True):
         """ Tag file/s with tag/s """
         # Parse arguments
-        file_names = []
+        file_paths = []
         tags = []
-        is_file_name = True
+        is_file_path = True
         for arg in args:
             if arg == "with":
-                is_file_name = False
+                is_file_path = False
                 continue
-            if is_file_name:
-                file_names.append(arg)
+            if is_file_path:
+                file_paths.append(arg)
             else:
-                tags.append(arg)
+                tag_val = arg.split("=")
+                if len(tag_val) == 1:
+                    tags.append((arg, None))
+                else:
+                    tags.append((tag_val[0], tag_val[-1]))
         if add:
-            self.add(*file_names)
-        # Add tags
-        for file_name in file_names:
-            file_name = file_name.split("/")[-1]
-            for tag in tags:
-                self.db.add_tag_to_file(tag, file_name)
-            # print("Tagged", file_name, "with", tags)
+            self.add(*file_paths)
+        # Add tags to files
+        for file_path in file_paths:
+            for tag, value in tags:
+                self.db.add_tag_to_file(tag, str(file_path), value)
+                parent_tag_ids = self.db.get_parent_tag_ids_by_name(tag)
+                # Add parent tags to file
+                for parent_tag_id in parent_tag_ids:
+                    self.db.add_tag_to_file(parent_tag_id, str(file_path))
         if commit:
             self.db.conn.commit()
         # Remount (everything is mounted)
@@ -373,7 +458,7 @@ class HyperTag:
             self.mount(self.root_dir)
 
     def metatags(self, *tag_names):
-        """ Display all metatags of tag/s """
+        """ Display all metatags (parents) of tag/s """
         tags = set()
         for tag_name in tag_names:
             tags.update(set(self.db.get_meta_tags_by_tag_name(tag_name)))
@@ -395,18 +480,28 @@ class HyperTag:
             else:
                 tags.append(arg)
 
-        # Add tags
+        # Add meta tags
         for tag in tags:
             for parent_tag in parent_tags:
                 self.db.add_parent_tag_to_tag(parent_tag, tag)
             # print("MetaTagged", tag, "with", parent_tags)
+
+            # Add tags to files
+            file_paths = self.db.get_files_by_tag(tag, show_path=True, fuzzy=True)
+            for file_path in file_paths:
+                self.db.add_tag_to_file(tag, str(file_path))
+                parent_tag_ids = self.db.get_parent_tag_ids_by_name(tag)
+                # Add parent tags to file
+                for parent_tag_id in parent_tag_ids:
+                    self.db.add_tag_to_file(parent_tag_id, str(file_path))
+
         if commit:
             self.db.conn.commit()
 
         for tag in tags:
             for parent_tag in parent_tags:
                 # Remove parent_tag dir in all levels
-                self.rmdir(self.root_dir, parent_tag)
+                remove_dir(self.root_dir, parent_tag)
                 # Remove tag dir in root level
                 try:
                     rmtree(self.root_dir / tag)
@@ -416,24 +511,16 @@ class HyperTag:
         if remount:
             self.mount(self.root_dir)
 
-    def rmdir(self, directory, del_dir_name):
-        # Delete dirs named del_dir_name in directory recursively
-        for item in directory.iterdir():
-            if item.name == del_dir_name and item.is_dir():
-                rmtree(item)
-            if item.is_dir():
-                self.rmdir(item, del_dir_name)
-
     def merge(self, tag_a, _into, tag_b):
         """ Merges all associations (files & tags) of tag_a into tag_b """
         print("Merging tag", tag_a, "into", tag_b)
         self.db.merge_tags(tag_a, tag_b)
-        self.rmdir(self.root_dir, tag_a)
+        remove_dir(self.root_dir, tag_a)
         self.mount(self.root_dir)
 
 
 def daemon():
-    """ Start daemon process watching HyperTagFS """
+    """ Start daemon process """
     print("Starting up daemon...")
     from .daemon import start
     from multiprocessing import Process, set_start_method
@@ -448,8 +535,10 @@ def help():
     """ Get some help on how to use HyperTag """
     print(
         """
+    README: https://github.com/SeanPedersen/HyperTag/blob/master/README.md
     Found a bug or simply feeling lost?
     Do not despair, you're not alone! Let us know at https://github.com/SeanPedersen/HyperTag/issues
+    Chat and read about HyperTag using Matrix at https://bit.ly/2Xtz2Ff
     Print available CLI functions: hypertag
     Print available CLI flags for all commands: hypertag command --help
     """
