@@ -4,15 +4,16 @@ from typing import List
 from pathlib import Path
 import filetype  # type: ignore
 from fuzzywuzzy import process  # type: ignore
+from .utils import is_int
 
 
 class Persistor:
     """ SQLite3 DB persistence class """
 
     def __init__(self):
-        db_path = Path.home() / ".config/hypertag/"
-        os.makedirs(db_path, exist_ok=True)
-        path = db_path / "hypertag.db"
+        self.db_path = Path.home() / ".config/hypertag/"
+        os.makedirs(self.db_path, exist_ok=True)
+        path = self.db_path / "hypertag.db"
         self.hypertagfs_dir = "hypertagfs_dir"
         self.hypertagfs_name = "HyperTagFS"
         self.ignore_list_name = "ignore_list"
@@ -99,7 +100,7 @@ class Persistor:
             """
         )
 
-        # Migrate old files table
+        # Migrate old files table TODO: Remove on 1.0 release (add to CREATE)
         try:
             self.c.execute("ALTER TABLE files ADD COLUMN embedding_vector TEXT")
         except sqlite3.OperationalError:
@@ -108,6 +109,10 @@ class Persistor:
             self.c.execute("ALTER TABLE files ADD COLUMN clean_text TEXT")
         except sqlite3.OperationalError:
             pass
+        try:
+            self.c.execute("ALTER TABLE files ADD COLUMN indexed INTEGER")
+        except sqlite3.OperationalError:
+            pass
 
         self.c.execute(
             """
@@ -131,6 +136,11 @@ class Persistor:
             )
             """
         )
+        # Migrate old tags_files table TODO: Remove on 1.0 release (add to CREATE)
+        try:
+            self.c.execute("ALTER TABLE tags_files ADD COLUMN value TEXT")
+        except sqlite3.OperationalError:
+            pass
 
         self.c.execute(
             """
@@ -154,6 +164,15 @@ class Persistor:
             )
             """
         )
+        try:
+            self.c.execute(
+                "ALTER TABLE auto_import_directories ADD COLUMN auto_index_images INTEGER"
+            )
+            self.c.execute(
+                "ALTER TABLE auto_import_directories ADD COLUMN auto_index_texts INTEGER"
+            )
+        except sqlite3.OperationalError:
+            pass
 
         for group, types in self.file_groups_types.items():
             self.add_tag(group)
@@ -237,10 +256,10 @@ class Persistor:
             (file_name, str(file_path)),
         )
         if len(file_name.split(".")) > 1 and file_type:
-            self.add_tag_to_file(file_type, file_name)
+            self.add_tag_to_file(file_type, str(file_path))
         file_group = self.file_types_groups.get(file_type)
         if file_group:
-            self.add_tag_to_file(file_group, file_name)
+            self.add_tag_to_file(file_group, str(file_path))
 
     def add_tag(self, name: str):
         self.c.execute(
@@ -255,20 +274,35 @@ class Persistor:
         tag_id = self.get_tag_id_by_name(name)
         return tag_id
 
-    def add_auto_import_directory(self, path: str):
+    def add_auto_import_directory(self, path: str, auto_index_images: bool, auto_index_text: bool):
+        path = str(Path(path))  # Standardizes path format (trailing / etc.)
         self.c.execute(
             """
-            INSERT OR IGNORE INTO auto_import_directories(
-                path
+            INSERT OR REPLACE INTO auto_import_directories(
+                path,
+                auto_index_images,
+                auto_index_texts
             )
-            VALUES(?)
+            VALUES(?, ?, ?)
             """,
-            [path],
+            [path, int(auto_index_images), int(auto_index_text)],
         )
         auto_import_dir_id = self.get_auto_import_id_by_path(path)
         self.conn.commit()
         return auto_import_dir_id
 
+    def get_auto_index_images(self, path):
+        self.c.execute(
+            "SELECT auto_index_images FROM auto_import_directories WHERE path LIKE ?", [path]
+        )
+        return bool(self.c.fetchone()[0])
+
+    def get_auto_index_texts(self, path):
+        self.c.execute(
+            "SELECT auto_index_texts FROM auto_import_directories WHERE path LIKE ?", [path]
+        )
+        return bool(self.c.fetchone()[0])
+
     def update_file_by_id(self, file_id: int, file_path: str):
         file_path = str(file_path)
         file_name = file_path.split("/")[-1]
@@ -317,7 +351,30 @@ class Persistor:
                     file_embedding_vectors.append((path, embedding_vector))
         return file_embedding_vectors
 
-    def get_indexed_file_paths(self, show_path=True):
+    def set_indexed_by_file_paths(self, file_paths: List[str]):
+        for file_path in file_paths:
+            self.c.execute(
+                """
+                UPDATE files
+                SET
+                    indexed = 1
+                WHERE
+                    path = ?
+                """,
+                [file_path],
+            )
+        self.conn.commit()
+
+    def get_unindexed_file_paths(self, show_path=True):
+        if show_path:
+            head = "SELECT path"
+        else:
+            head = "SELECT name"
+        self.c.execute(head + " FROM files WHERE indexed is NULL")
+        data = [e[0] for e in self.c.fetchall()]
+        return data
+
+    def get_vectorized_file_paths(self, show_path=True):
         if show_path:
             head = "SELECT path"
         else:
@@ -326,7 +383,7 @@ class Persistor:
         data = [e[0] for e in self.c.fetchall()]
         return data
 
-    def get_unindexed_file_paths(self) -> List[str]:
+    def get_unvectorized_file_paths(self) -> List[str]:
         self.c.execute(
             "SELECT path FROM files WHERE embedding_vector is NULL AND clean_text is NULL"
         )
@@ -341,19 +398,27 @@ class Persistor:
         data = [e[0] for e in self.c.fetchall()]
         return data
 
-    def add_tag_to_file(self, tag_name: str, file_name: str):
-        file_id = self.get_file_id_by_name(file_name)
-        tag_id = self.add_tag(tag_name)
+    def add_tag_to_file(self, tag_name_or_id: str, file_name_or_path: str, value=None):
+        if is_int(tag_name_or_id):
+            tag_id = int(tag_name_or_id)
+        else:
+            tag_id = self.add_tag(tag_name_or_id)
+
+        if "/" in file_name_or_path:
+            file_id = self.get_file_id_by_path(file_name_or_path)
+        else:
+            file_id = self.get_file_id_by_name(file_name_or_path)
 
         self.c.execute(
             """
             INSERT OR IGNORE INTO tags_files(
                 file_id,
-                tag_id
+                tag_id,
+                value
             )
-            VALUES(?, ?)
+            VALUES(?, ?, ?)
             """,
-            (file_id, tag_id),
+            (file_id, tag_id, value),
         )
 
     def remove_tag_from_file(self, tag_name: str, file_name: str):
@@ -373,6 +438,16 @@ class Persistor:
         self.c.execute("SELECT tag_id FROM tags WHERE name LIKE ?", [name])
         return self.c.fetchone()[0]
 
+    def get_parent_tag_ids_by_name(self, tag_name: str):
+        self.c.execute(
+            """
+        SELECT tt.parent_tag_id
+        FROM tags as t, tags_tags as tt
+        WHERE tt.children_tag_id = t.tag_id AND name LIKE ?""",
+            [tag_name],
+        )
+        return [e[0] for e in self.c.fetchall()]
+
     def get_file_id_by_path(self, path: str):
         self.c.execute("SELECT file_id FROM files WHERE path LIKE ?", [path])
         path = self.c.fetchone()
@@ -581,7 +656,8 @@ class Persistor:
         data = [e[0] for e in self.c.fetchall()]
         return data
 
-    def get_file_paths_names_by_tag_id(self, tag_id):
+    def get_file_paths_names_by_tag_id_shallow(self, tag_id):
+        # Includes all files of tag id
         self.c.execute(
             """
             SELECT f.path, f.name
@@ -595,7 +671,26 @@ class Persistor:
         data = self.c.fetchall()
         return data
 
-    def get_files_by_tag(self, tag_name, show_path, fuzzy, verbose=False):
+    def get_file_paths_names_by_tag_id(self, tag_id, files_data=[]):
+        # Recursively includes all children tags files
+        self.c.execute(
+            """
+            SELECT f.path, f.name
+            FROM files f, tags t, tags_files tf
+            WHERE f.file_id = tf.file_id AND
+                tf.tag_id = t.tag_id AND
+                t.tag_id = ?
+            """,
+            [tag_id],
+        )
+        data = self.c.fetchall() + files_data
+
+        children_tuples = self.get_tag_id_children_ids_names(tag_id)
+        for child_tag_id, _name in children_tuples:
+            data = self.get_file_paths_names_by_tag_id(child_tag_id, data)
+        return data
+
+    def get_files_by_tag(self, tag_name, show_path, fuzzy, value=None, verbose=False):
         if fuzzy:
             matches = process.extract(tag_name, self.get_tags(), limit=5)
             best_match = None, None
@@ -615,6 +710,11 @@ class Persistor:
             select = "SELECT f.path"
         else:
             select = "SELECT f.name"
+        value_q = ""
+        bindings = [tag_name]
+        if value:
+            value_q = "AND tf.value LIKE ?"
+            bindings = [tag_name, value]
         self.c.execute(
             select
             + """
@@ -622,8 +722,9 @@ class Persistor:
             WHERE f.file_id = tf.file_id AND
                 tf.tag_id = t.tag_id AND
                 t.name LIKE ?
-            """,
-            [tag_name],
+            """
+            + value_q,
+            bindings,
         )
         data = [e[0] for e in self.c.fetchall()]
         return data
